/* ToBe Pilot
 * Inputs:
 *   - User buttons, Power buttons, Over Current status
 *   - Call userspace script.
 *
 * Outputs:
 *   - Power enables, LEDs, control GPIO (reset, boot)
 *   - controlled via /sys/devices pseudo filesystem
 *
 * UART:
 *   Uart uses Rpi 3/ Zero-W mini UART, if Bluetooth not needed, this can be changed
 *   add 'dtoverlay=pi3-disable-bt' to /boot/config.txt
 *   Linux console uses serial0, to remove: sudo nano /boot/cmdline.txt
 *     remove : console=serial0,115200
 *
 * Automatically load at boot:
 *   sudo make install
 *   sudo depmod
 *   add tobe-pilot to /etc/modules
 *
 * I2C/SPI interfaces, when needed, can be enabled via Device Tree
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kobject.h>

#include "pilot.h"
#include "rpi.h"
#include "tick.h"
#include "input.h"
#include "output.h"

// SysFs support --------------------------------------------------------------

static ssize_t statsShow(struct device *dev, struct device_attribute *attr, char *buf) {
  int nb;

  nb  = outPrintStats(buf, PAGE_SIZE);
  nb += inPrintStats(buf+nb, PAGE_SIZE-nb);
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdate-time"
  nb += scnprintf(buf+nb, PAGE_SIZE-nb, "ToBe Pilot:\n  hw-rev: %s\n  sw-ver: %s (built " __DATE__ ")\n",
        pilots[pilotsIdx].version, THIS_MODULE->version);
#pragma GCC diagnostic pop
  nb += scnprintf(buf+nb, PAGE_SIZE-nb, "Raspberry Pi %s:\n  %s: %s %s, %s\n  board-rev: %s (%s)\n",
        rpiGetType(), rpiGetProcessor(), rpiGetArmPart(), rpiGetArmRevision(), rpiGetMemorySize(),
        rpiGetBoardRevision(), rpiGetManufacturer());
  return nb;
}

static ssize_t statsStore(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
  if (sysfs_streq("reset", buf))
    inResetStats();
  return count;
}

//static DEVICE_ATTR(stats, 0664, statsShow, statsStore);
static struct device_attribute statsDevAttr = {
  .attr = {
    .name = "stats",
    .mode = S_IWUSR | S_IRUGO,
  },
  .show = statsShow,
  .store = statsStore,
};

static enum { IDLE, RUN } status = IDLE;

static ssize_t statusShow(struct device *dev, struct device_attribute *attr, char *buf) {
  return scnprintf(buf, PAGE_SIZE, "%s\n", status == RUN ? "run" : "idle");
}

static ssize_t statusStore(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
  if (sysfs_streq("0", buf) || sysfs_streq("reset", buf)) {
    outLedSet("red", OFF);
    outLedSet("green", OFF);
    status = IDLE;
  } else if (sysfs_streq("run", buf)) {
    outLedSet("red", OFF);
    outLedSet("green", BLINK);
    status = RUN;
  } else if (sysfs_streq("ok", buf)) {
    outLedSet("red", OFF);
    outLedSet("green", FLASH_ONCE);
    status = IDLE;
  } else if (sysfs_streq("err", buf)) {
    outLedSet("green", OFF);
    outLedSet("red", BLINK);
    status = IDLE;
  }
  return count;
}

static struct device_attribute statusDevAttr = {
  .attr = {
    .name = "status",
    .mode = S_IWUSR | S_IRUGO,
  },
  .show = statusShow,
  .store = statusStore,
};

// sysfs data -----------------------------------------------------------------
static struct device *dev = NULL;
static struct kobject *obj = NULL;

static struct attribute *attrs[] = { &statsDevAttr.attr, &statusDevAttr.attr, NULL };
static struct attribute_group attrGroup = { .attrs = attrs };

// create /sys/devices/tobe-pilot structure
static int sysfsInit(void) {
  // create device
  dev = root_device_register("tobe");
  // create kObject
  if ( (obj = kobject_create_and_add("pilot", &(dev->kobj))) == 0 ) {
    printk(KERN_WARNING "ToBe-Pilot: Can't init pilot kObj.\n");
    return -ENOMEM;
  }

  // create sysfs entries
  if (sysfs_create_group(obj, &attrGroup)) {
    printk(KERN_WARNING "ToBe-Pilot: Can't create pilot sysfs group.\n");
    return -ENOMEM;
  }

  outCreateSysFs(obj);
  return 0;
}

// ----------------------------------------------------------------------------
// Modules parameters
// Note: must be set at startup, dynamic change is not supported
// sudo insmod hwrev=2.0
static char *hwrev = "v1.1";
module_param(hwrev, charp, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(hwrev, "Hardware revision (default=v1.1)");
// ----------------------------------------------------------------------------

static int __init pilotInit(void) {
  int i;

  setHardware(hwrev);
  printk(KERN_INFO "ToBe-Pilot(hw:%s): Initialising RPi %s (%s).\n",
         pilots[pilotsIdx].version, rpiGetType(), rpiGetProcessor());
  tickInit();  // Periodic polling

  // Outputs init
  outInit();
  for (i=0; i<pilots[pilotsIdx].outsNb; i++) {
    outAdd(&(pilots[pilotsIdx].outs[i]));
  }

  // Enable pull ups
  if (inInit(pilots[pilotsIdx].pullups))
    printk(KERN_WARNING "ToBe-Pilot: Pull-ups init failed.\n");

  // Inputs init
  for (i=0; i<pilots[pilotsIdx].insNb; i++)
    inAdd(&(pilots[pilotsIdx].ins[i]));

  // Sysfs
  if (sysfsInit())
    printk(KERN_WARNING "ToBe-Pilot: /sys/device/tobe-pilot init failed.\n");

  // call userspace init
  inScriptRun("pilot init", 1);

  return 0;
}

static void __exit pilotExit(void) {
  tickExit();
  inExit();
  outExit(obj);

  if (obj) {
    sysfs_remove_group(obj, &attrGroup);
    kobject_put(obj);
  }
  if (dev) root_device_unregister(dev);

  printk(KERN_INFO "ToBe-Pilot: Leaving.\n");
}

module_init(pilotInit);
module_exit(pilotExit);

MODULE_DESCRIPTION("ToBe Pilot driver");
MODULE_VERSION("1.1-x");
MODULE_AUTHOR("Bertrand Tognoli");
MODULE_LICENSE("GPL");
