/** Raspberry Pi identifcation utils.
 * Info obtained from:
 * <a href="https://github.com/raspberrypi/documentation/blob/master/hardware/raspberrypi/revision-codes/README.md">https://github.com/raspberrypi/documentation/blob/master/hardware/raspberrypi/revision-codes/README.md</a>
 *
 * @code
 * #include "rpi.h"
 * extern unsigned int system_rev;
 *
 * void dumpInfo(void {
 *   printk(" RPi %s (%x): %s, %s %s, %s, Board-Rev %s (%s)\n",
 *     rpiGetType(), system_rev,
 *     rpiGetProcessor(), rpiGetArmPart(), rpiGetArmRevision(), rpiGetMemorySize(),
 *     rpiGetBoardRevision(), rpiGetManufacturer());
 * }
 * @endcode
 */

#include "rpi.h"

#define sizeofarray(a) (sizeof(a)/sizeof(a[0]))

// Arm Kernel exports system revision:
extern unsigned int system_rev;

static const char *PROCS[] = {"BCM2835", "BCM2836", "BCM2837"};
static const char *TYPES[] = {"A", "B", "A+", "B+", "2B", "Alpha", "CM1", "?", "3B", "Zero", "CM3", "?", "Zero W", "3B+"};
static unsigned MEM_SZ[] = {256, 512, 1024};
static const char *MANUFS[] = {"Sony UK", "Egoman", "Embest", "Sony Japan", "Embest", "Stadium" };

static const struct {
  const char * type;
  const char * rev;
  unsigned mem;
  const char * manuf;
} OLDS[] = {
  {"?",   "1.0",   0, "?"},
  {"?",   "1.0",   0, "?"},
  {"B",   "1.0", 256, "Egoman"},
  {"B",   "1.0", 256, "Egoman"},
  {"B",   "2.0", 256, "Sony UK"},
  {"B",   "2.0", 256, "Qisda"},
  {"B",   "2.0", 256, "Egoman"},
  {"A",   "2.0", 256, "Egoman"},
  {"A",   "2.0", 256, "Sony UK"},
  {"A",   "2.0", 256, "Qisda"},
  {"?",   "1.0",   0, "?"},
  {"?",   "1.0",   0, "?"},
  {"?",   "1.0",   0, "?"},
  {"B",   "2.0", 512, "Egoman"},
  {"B",   "2.0", 512, "Sony UK"},
  {"B",   "2.0", 512, "Egoman"},
  {"B+",  "1.0", 512, "Sony UK"},
  {"CM1", "1.0", 512, "Sony UK"},
  {"A+",  "1.1", 256, "Sony UK"},
  {"B+",  "1.2", 512, "Embest"},
  {"CM1", "1.0", 512, "Embest"},
  {"A+",  "1.1", 512, "Embest"},
};

// xxxxxxxx 1xxxxxxx PPPPxxxx xxxxxxxx
const char *rpiGetProcessor(void) {
  unsigned revId = system_rev;
  if (revId & 0x00800000) {
    revId = (revId>>12) & 0xf;
    if (revId >= sizeofarray(PROCS)) return "???";
    return PROCS[revId];
  }
  return PROCS[0];
}

// xxxxxxxx 1xxxxxxx xxxxTTTT TTTTxxxx
const char *rpiGetType(void) {
  unsigned revId = system_rev;
  if (revId & 0x00800000) {
    revId = (revId>>4) & 0xff;
    if (revId >= sizeofarray(TYPES)) return "?";
    return TYPES[revId];
  }
  if (revId >= sizeofarray(OLDS)) return "?";
  return OLDS[revId].type;
}

// xxxxxxxx 1xxxxxxx xxxxxxxx xxxxRRRR
static char boardRev[4] = "1.0";
const char *rpiGetBoardRevision(void) {
  unsigned revId = system_rev;
  if (revId & 0x00800000) {
    revId &= 0xf;
    if (revId < 10)
      boardRev[2] = '0'+revId;
    else
      boardRev[2] = 'a' - 10 + revId;
    return boardRev;
  }
  if (revId >= sizeofarray(OLDS)) return "?";
  return OLDS[revId].rev;
}

// xxxxxxxx 1xxxMMMM xxxxxxxx xxxxxxxx
const char *rpiGetManufacturer(void) {
  unsigned revId = system_rev;
  if (revId & 0x00800000) {
    revId = (revId>>16) & 0xf;
    if (revId >= sizeofarray(MANUFS)) return "?";
    return MANUFS[revId];
  }
  if (revId >= sizeofarray(OLDS)) return "?";
  return OLDS[revId].manuf;
}

// xxxxxxxx 1MMMxxxx xxxxxxxx xxxxxxxx
unsigned rpiGetMemorySizeInMB(void) {
  unsigned revId = system_rev;

  if (revId & 0x00800000) {
    revId = (revId>>20) & 0x7;
    if (revId >= sizeofarray(MEM_SZ)) return 0;
    return MEM_SZ[revId];
  }
  if (revId >= sizeofarray(OLDS)) return 0;
  return OLDS[revId].mem;
}

const char *rpiGetMemorySize(void) {
  switch (rpiGetMemorySizeInMB()) {
    case 256: return "256MB";
    case 512: return "512MB";
    case 1024: return "1GB";
  }
  return "?MB";
}

int rpiIsBCM2835(void) {
  unsigned revId = system_rev;
  if (revId & 0x00800000)
    return ((revId>>12) & 0xf) == 0;
  return 1;
}

unsigned rpiGetArmMainIdReg(void) {
  unsigned id = 0;
#ifdef __arm__
  asm volatile ("MRC p15, 0, %0, c0, c0, 0" : "=r"(id));
#endif
  return id;
}

// IIIIIIII xxxxxxxx xxxxxxxx xxxxxxxx Implementer (0x41: ARM)
// xxxxxxxx VVVVxxxx xxxxxxxx xxxxxxxx Variant Major
// xxxxxxxx xxxxAAAA xxxxxxxx xxxxxxxx Architecture
// xxxxxxxx xxxxxxxx PPPPPPPP PPPPxxxx Part Nb (0xC09 - CortexA9)
// xxxxxxxx xxxxxxxx xxxxxxxx xxxxRRRR Revision Minor

const char * rpiGetArmPart(void) {
  switch ((rpiGetArmMainIdReg() >> 4) & 0xfff) {
    case 0xb76: return "ARM1176JZF-S";
    case 0xc07: return "Cortex-A7";
    case 0xc09: return "Cortex-A9";
    case 0xd03: return "Cortex-A53";
  }
  return "Unknown";
}

static char armRev[5] = "r0p0";

const char * rpiGetArmRevision(void) {
  unsigned midr = rpiGetArmMainIdReg();
  unsigned char major = (midr>>20) & 0xf;
  unsigned char minor = midr & 0xf;

  armRev[1] = (major < 10 ? '0'+major : 'A'-10+major);
  armRev[3] = (minor < 10 ? '0'+minor : 'A'-10+minor);
  return armRev;
}

#ifndef __arm__
unsigned int system_rev = 0; // to get rid of make warning...
#endif
